@@ -11,6 +11,7 @@ use I18N::Langinfo qw/langinfo/;
 use List::Util qw/max/;
 
 our $VERSION = '0.0.1';
+our $COLOUR  = 0;
 
 my $CODESET = eval { I18N::Langinfo::CODESET() };
 my %seen;           # hash of hashes keyed by 'caller', then opt/arg name
@@ -223,79 +224,151 @@ sub arg {
 # ------------------------------------------------------------------------
 
 sub _usage {
-    my $caller = shift;
-    my $error  = shift;
-
-    my @config;
+    my $caller   = shift;
+    my $error    = shift;
+    my $terminal = -t STDOUT;
+    my $red      = $terminal ? '[0;31m' : '';
+    my $yellow   = $COLOUR && $terminal ? '[0;33m' : '';
+    my $grey     = $COLOUR && $terminal ? '[1;30m' : '';
+    my $reset    = $COLOUR && $terminal ? '[0m' : '';
     my $parent   = $caller;
-    my $length_a = 0;
-    my $length_b = 0;
-    my $usage    = '';
+    my @args     = @{ $args{$caller} };
+    my @opts     = @{ $opts{$caller} };
+    my @parents;
+    my @usage;
+    my @uargs;
+    my @uopts;
+    my $usage;
 
-    foreach my $def ( @{ $args{$caller} } ) {
-        $usage .= ' ' if $usage;
-        $usage .= '[' unless $def->{required};
-        $usage .= uc $def->{name};
-        $usage .= '...' if $def->{greedy};
-        $usage .= ']' unless $def->{required};
+    require File::Basename;
+    my $me = File::Basename::basename($0);
 
-        $length_a = max( $length_a, map { length $_ } @{ $def->{subcommands} } )
-          if $def->{isa} eq 'SubCmd';
-    }
+    $usage .= $yellow . 'usage:' . $reset;
 
     while ( $parent =~ s/(.*)::(.*)/$1/ ) {
         last unless $seen{$parent};
         ( my $name = $2 ) =~ s/_/-/g;
-        $usage = $name . ' ' . $usage;
-        unshift( @config, @{ $opts{$parent} } );
+        unshift( @parents, $name );
+        unshift( @opts,    @{ $opts{$parent} } );
     }
 
-    $usage .= "\n";
-    unshift( @config, @{ $args{$caller} } );
-    push( @config, @{ $opts{$caller} } );
+    $me .= ' ' . join( ' ', @parents ) if @parents;
 
-    $length_b =
-      max( map { $_->{length} + 4 } grep { $_->{type} eq 'opt' } @config ) || 0;
-    $length_a = max( $length_a + 4, $length_b + 5 ) || 0;
-
-    my $format_a  = '    %-' . $length_a . "s   %-s\n";
-    my $format_b  = '%-' . $length_b . 's%-2s';
-    my $prev_type = '';
-
-    foreach my $def (@config) {
-        $usage .= "\n" if $def->{type} ne $prev_type;
-        $prev_type = $def->{type};
-
-        if ( $def->{type} eq 'arg' ) {
-            $usage .= sprintf( $format_a, uc $def->{name}, $def->{comment} );
-            foreach my $subcommand ( @{ $def->{subcommands} } ) {
-                my $pkg = $def->{package} . '::' . $subcommand;
+    if ( my $last = $args[$#args] ) {
+        if ( $last->{isa} eq 'SubCmd' ) {
+            foreach my $subcommand ( @{ $last->{subcommands} } ) {
+                my $pkg = $last->{package} . '::' . $subcommand;
                 $pkg =~ s/-/_/g;
-                my $desc = $desc{$pkg};
-                $usage .= sprintf( $format_a, '  ' . $subcommand, $desc );
+                push( @usage, [ $me . ' ' . $subcommand, $desc{$pkg} ] );
             }
 
-            if ( $def->{fallback} ) {
-                $usage .= sprintf( $format_a,
-                    '  ' . uc $def->{fallback}->{name},
-                    $def->{fallback}->{comment} );
+            if ( $last->{fallback} ) {
+                push(
+                    @usage,
+                    [
+                        $me . ' ' . uc $last->{fallback}->{name},
+                        $last->{fallback}->{comment}
+                    ]
+                );
             }
+            $usage .= "\n";
         }
-        elsif ( $def->{type} eq 'opt' ) {
-            my $opt = '--' . $def->{name};
-            $opt =~ s/_/-/g;
-            $opt .= ',' if $def->{alias};
-            my $tmp = sprintf( $format_b,
-                $opt, $def->{alias} ? '-' . $def->{alias} : '' );
-
-            $usage .= sprintf( $format_a, $tmp, $def->{comment} );
+        else {
+            $usage .= ' ' . $me;
+            foreach my $def (@args) {
+                $usage .= ' ';
+                $usage .= '[' unless $def->{required};
+                $usage .= uc $def->{name};
+                $usage .= '...' if $def->{greedy};
+                $usage .= ']' unless $def->{required};
+                push( @uargs, [ uc $def->{name}, $def->{comment} ] );
+            }
+            $usage .= "\n";
         }
     }
+    else {
+        $usage .= ' ' . $me . "\n";
+    }
+
+    foreach my $opt (@opts) {
+        ( my $name = $opt->{name} ) =~ s/_/-/g;
+        $name .= ',' if $opt->{alias};
+        push(
+            @uopts,
+            [
+                '--' . $name,
+                $opt->{alias}
+                ? '-' . $opt->{alias}
+                : '',
+                $opt->{comment}
+            ]
+        );
+    }
+
+    if (@uopts) {
+        my $w1 = max( map { length $_->[0] } @uopts );
+        my $fmt = '%-' . $w1 . "s %s";
+
+        @uopts = map { [ sprintf( $fmt, $_->[0], $_->[1] ), $_->[2] ] } @uopts;
+    }
+
+    my $w1 = max( map { length $_->[0] } @usage, @uargs, @uopts );
+    my $format = '    %-' . $w1 . "s   %s\n";
+
+    if (@usage) {
+        foreach my $row (@usage) {
+            $usage .= sprintf( $format, @$row );
+        }
+    }
+    if (@uargs) {
+        $usage .= "\n  ${grey}arguments:$reset\n";
+        foreach my $row (@uargs) {
+            $usage .= sprintf( $format, @$row );
+        }
+    }
+    if (@uopts) {
+        $usage .= "\n  ${grey}options:$reset\n";
+        foreach my $row (@uopts) {
+            $usage .= sprintf( $format, @$row );
+        }
+    }
+
+    if ($error) {
+        my $tmp = $red . "error:" . $reset;
+        $tmp .= "\n   " if @usage;
+        return $tmp . ' ' . $error . "\n\n" . $usage . "\n";
+    }
+    return $usage . "\n";
+}
+
+sub _synopsis {
+    my $caller = shift;
+    my $parent = $caller;
+    my @args   = @{ $args{$caller} };
+    my @parents;
 
     require File::Basename;
-    $usage = 'usage: ' . File::Basename::basename($0) . ' ' . $usage . "\n";
-    $usage = $error . "\n\n" . $usage if $error;
-    return $usage;
+    my $usage = File::Basename::basename($0);
+
+    while ( $parent =~ s/(.*)::(.*)/$1/ ) {
+        last unless $seen{$parent};
+        ( my $name = $2 ) =~ s/_/-/g;
+        unshift( @parents, $name );
+    }
+
+    $usage .= ' ' . join( ' ', @parents ) if @parents;
+
+    if ( my $last = $args[$#args] ) {
+        foreach my $def (@args) {
+            $usage .= ' ';
+            $usage .= '[' unless $def->{required};
+            $usage .= uc $def->{name};
+            $usage .= '...' if $def->{greedy};
+            $usage .= ']' unless $def->{required};
+        }
+    }
+
+    return 'usage: ' . $usage . "\n";
 }
 
 sub usage {
@@ -312,14 +385,20 @@ sub _optargs {
     my $package = $caller;
 
     if ( !@_ and @ARGV ) {
+        require Test::More;
+        use Data::Dumper;
+        Test::More::diag(
+            Dumper(
+                [ @ARGV, 'codeset:', $CODESET, eval { langinfo($CODESET) } ]
+            )
+        );
 
         if ($CODESET) {
             my $codeset = langinfo($CODESET);
-            @ARGV =
-              map { utf8::is_utf8($_) ? $_ : decode( $codeset, $_ ) } @ARGV;
+            $_ = decode( $codeset, $_ ) for @ARGV;
         }
         else {
-            @ARGV = map { utf8::is_utf8($_) ? $_ : decode_utf8($_) } @ARGV;
+            $_ = decode( 'UTF-8', $_ ) for @ARGV;
         }
 
         $source = \@ARGV;
@@ -327,9 +406,9 @@ sub _optargs {
 
     croak "no option or argument defined for $caller"
       unless exists $opts{$package}
-          or exists $args{$package};
+      or exists $args{$package};
 
-    Getopt::Long::Configure(qw/pass_through no_auto_abbrev/);
+    Getopt::Long::Configure(qw/pass_through no_auto_abbrev no_ignore_case/);
 
     my @config = ( @{ $opts{$package} }, @{ $args{$package} } );
 
@@ -345,6 +424,9 @@ sub _optargs {
         }
         elsif ( $try->{type} eq 'arg' ) {
             if (@$source) {
+                die _usage( $package, "unknown option: " . $source->[0] )
+                  if ( $source->[0] =~ m/^(-\S)|(--\S+)$/ );
+
                 if ( $try->{greedy} ) {
                     my @later;
                     if ( @config and @$source > @config ) {
@@ -384,9 +466,6 @@ sub _optargs {
             }
 
             if ( $try->{isa} eq 'SubCmd' and $result ) {
-                die _usage( $package, "unknown option: " . $result )
-                  if ( $result =~ m/^-/ );
-
                 my $newpackage = $package . '::' . $result;
                 $newpackage =~ s/-/_/;
 
@@ -396,7 +475,7 @@ sub _optargs {
                 }
                 elsif ( !$try->{fallback} and !$ishelp ) {
                     die _usage( $package,
-                        "unknown " . uc( $try->{name} ) . ': ' . $result );
+                        "invalid " . $try->{name} . ': ' . $result );
                 }
             }
 
